import * as superagent from 'superagent'
import * as firebase from "@firebase/testing";
import {WhereFilterOp} from "@firebase/firestore-types";
import {FirestoreClient} from "../FirestoreClient"
import * as Firebase from "@firebase/testing";
import { DocumentData } from "@google-cloud/firestore";


const projectId = "ocf-integration-test"
const host = `http://localhost:5001/${projectId}/us-central1`
const crudFunctionName = "test_trigger_http_fs_v1 "

const collectionName = 'test'
let collection
let firestore
let fireStoreClient : FirestoreClient

describe('Integration Tests (offline using emulators)', () => {

    afterEach(async () => {
        await firebase.clearFirestoreData({projectId: projectId});
        firebase.apps().forEach(async app => await app.delete());
    });

    beforeEach(async () => {
        firestore = firebase.initializeTestApp({
            projectId: projectId
        }).firestore();
        fireStoreClient = new FirestoreClient(firestore)

        collection = firestore.collection(
            collectionName
        );
    })

    it('create doc, update it and check version update', async () => {
        //given
        const entityForCreation = {someMandatoryAttrib: "someVal", someOptionalAttrib: "someOptionalVal"} as EntityForCreation

        //when
        let response = await addDocument(entityForCreation)

        //then
        expect(response).toBeDefined()
        let entity : Entity = response.body
        expect(entity.created).toBeDefined()
        expect(entity.lastModified).toBeDefined()
        expect(entity.version).toBe(1)

        //and the entity is updated
        entity.someMandatoryAttrib = "someOtherValue"
        delete entity.someOptionalAttrib

        response = await setDocument(entity.id, entity, ["someOptionalAttrib"])

        //then
        expect(response).toBeDefined()
        entity = response.body
        expect(entity.created).toBeDefined()
        expect(entity.lastModified).toBeDefined()
        expect(entity.version).toBe(2)
        expect(entity.someMandatoryAttrib).toBe("someOtherValue")
        expect(entity.someOptionalAttrib).toBeUndefined()

    })

    it('create doc, delete and checkIfExists', async () => {
        //given
        const entityForCreation = {someMandatoryAttrib: "someVal", someOptionalAttrib: "someOptionalVal"} as EntityForCreation

        //when
        let response = await addDocument(entityForCreation)

        //then
        expect(response).toBeDefined()
        let entity : Entity = response.body
        expect(await documentExists(entity.id)).toBeTruthy

        //and the entity is deleted
        response = await deleteDocument(entity.id)

        //then
        expect(await documentExists(entity.id)).toBeFalsy
    })


    it('bulk import of 1002 docs and execute query', async () => {
        //given
        const data = [] as EntityForCreation[]

        for (let i = 0; i < 1002; i++) {
            if (i == 1 || i == 5 || i == 10) {
                data.push({someMandatoryAttrib: "someFilteredVal", someOptionalAttrib: "someOptionalVal"} as EntityForCreation)
            } else {
                data.push({someMandatoryAttrib: "someVal", someOptionalAttrib: "someOptionalVal"} as EntityForCreation)
            }
        }

        //when
        await bulkSet(data)
        const response = await executeQuery("someMandatoryAttrib", "==", "someVal")


        //then
        expect(response).toBeDefined()
    })

    it('create doc, than update within transaction', async () => {
        //given
        const entityForCreation = {someMandatoryAttrib: "someVal", someOptionalAttrib: "someOptionalVal"} as EntityForCreation
        let response = await addDocument(entityForCreation)
        let entity : Entity = response.body
        expect(entity.version).toBe(1)

        //when
        const result = await updateIdInTransaction(entity.id)

        //then
        response = await getDocument(entity.id)
        entity  = response.body
        expect(entity.version).toBe(2)
    })

    it('create docs and receive in bulk', async () => {
        //given
        const entityForCreation = {someMandatoryAttrib: "someVal", someOptionalAttrib: "someOptionalVal"} as EntityForCreation
        let response = await addDocument(entityForCreation)
        let entity : Entity = response.body
        expect(entity.version).toBe(1)

        const entityForCreation2 = {someMandatoryAttrib: "someVal", someOptionalAttrib: "someOptionalVal"} as EntityForCreation
        response = await addDocument(entityForCreation)
        let entity2 : Entity  = response.body
        expect(entity2.version).toBe(1)

        //when
        const result = await getDocumentsById([entity.id, entity2.id])

        //then
        let docs : Entity[]  = result.body
        expect(docs.length).toBe(2)
    })

    it('prepare collection entry, create doc, update and check on subcollection', async () => {
        //given
        const entityForCreation = {
            someMandatoryAttrib: "someVal",
            someOptionalAttrib: "someOptionalVal"
        } as EntityForCreation

        //when
        let response = await addDocument(entityForCreation)

        //then
        expect(response).toBeDefined()
        let entity : Entity = response.body
        expect(await documentExists(entity.id)).toBeTruthy
        const originalDocId = entity.id


        //and create entry in subcollection
        const entityForCreationSubcollection = {
            someSubCollAttrib: "someVal",
            someOptionalSubCollAttrib: "someOptionalVal"
        }

        //when
        response = await addDocument(entityForCreationSubcollection, "subCollectionName", originalDocId)

        //then
        expect(response).toBeDefined()
        console.log(response.body)
        entity = response.body
        expect(await documentExists(entity.id, "subCollectionName", originalDocId)).toBeTruthy


        //and
        response = await listDocuments(originalDocId,"subCollectionName")

        //then
        expect(response).toBeDefined()
        expect(response.body.length).toBe(1)
        expect(response.body[0].id).toBe(entity.id)


    })


    it('create doc, update and check', async () => {
        //given
        const entityForCreation = {someMandatoryAttrib: "someVal", someOptionalAttrib: "someOptionalVal"} as EntityForCreation

        //when
        let response = await addDocument(entityForCreation)

        //then
        expect(response).toBeDefined()
        let entity : Entity = response.body
        expect(await documentExists(entity.id)).toBeTruthy

        //and the entity is deleted
        response = await updateDocument(entity.id, {someOptionalAttrib: "someOtherOptionalVal"})
        expect(response.body.someOptionalAttrib).toBe("someOtherOptionalVal")
        expect(response.body.version).toBe(2)

    })

    it('check proper date transformation', async () => {
        //given
        let d = "2020-04-10T08:50:00.000Z"
        const entityForCreation = {
            someMandatoryAttrib: "someVal",
            someOptionalAttrib: "someOptionalVal",
            someOptionalDate: d} as unknown as EntityForCreation

        //when
        let response = await addDocument(entityForCreation)


        //then
        expect(response).toBeDefined()
        expect(response.body.someOptionalDate).toBe("2020-04-10T08:50:00.000Z")
    })

    it('check proper date transformation2', async () => {
        //given
        let d = new Date()
        d.setUTCFullYear(2020)
        d.setUTCMonth(1)
        d.setUTCDate(1)
        d.setUTCHours(10)
        d.setUTCMinutes(1)
        d.setUTCSeconds(1)
        d.setUTCMilliseconds(100)
        const entityForCreation = {
            someMandatoryAttrib: "someVal",
            someOptionalAttrib: "someOptionalVal",
            someOptionalDate: d} as EntityForCreation

        //when
        let response = await addDocument(entityForCreation)


        //then
        expect(response).toBeDefined()
        expect(response.body.someOptionalDate).toBe("2020-02-01T10:01:01.100Z")

        // and

        const docRef = collection.doc(response.body.id)
        const dataSnapshot = await new Promise<DocumentData>(
            (resolve, reject) => {
                docRef.onSnapshot((snap) => {
                    if (snap.empty) {
                        console.log("No matching documents.");
                        return;
                    }
                    resolve(snap);
                });
            }
        );

        expect(dataSnapshot.data().someOptionalDate.seconds).toBeDefined()
        expect(dataSnapshot.data().someOptionalDate.nanoseconds).toBeDefined()


    })


    it('bulk import of 20 docs and execute query with multiple conditions', async () => {
        //given
        const data = [] as EntityForCreation[]

        for (let i = 0; i < 20; i++) {
            if (i == 1 || i == 5 || i == 10) {
                data.push({someMandatoryAttrib: "someFilteredVal", someOptionalAttrib: "someOptionalVal1"} as EntityForCreation)
            } else if (i == 2 || i == 6 || i == 12) {
                data.push({someMandatoryAttrib: "someVal", someOptionalAttrib: "someOptionalVal1"} as EntityForCreation)
            } else {
                data.push({someMandatoryAttrib: "someVal", someOptionalAttrib: "someOptionalVal3"} as EntityForCreation)
            }
        }

        //when single query
        await bulkSet(data)
        const response = await executeQuery("someMandatoryAttrib", "==", "someVal")

        //then
        expect(response).toBeDefined()
        expect(response.body.length).toBe(17)

        // and

        //when multi condition query
        const responseMultiCondition = await getQuery(
            [
                {
                    fieldPath: "someMandatoryAttrib",
                    opStr: "==",
                    value: "someVal"
                },
                {
                    fieldPath: "someOptionalAttrib",
                    opStr: "==",
                    value: "someOptionalVal1"
                },
            ]
        )
        //then
        expect(responseMultiCondition).toBeDefined()
        expect(responseMultiCondition.body.length).toBe(3)


    })

})


async function executeQuery(fieldPath: string, opStr: WhereFilterOp, value: any, subCollection?: string, documentId?: string) {
    return await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            methodName: "executeQuery",
            args: [fieldPath, opStr, value],
            subCollection: subCollection,
            documentId: documentId
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
}

async function getQuery(queries : {fieldPath: string, opStr: WhereFilterOp, value: any}[], subCollection?: string, documentId?: string) {
    return await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            methodName: "getQuery",
            args: [queries],
            subCollection: subCollection,
            documentId: documentId
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
}

async function updateIdInTransaction(id: string, subCollection?: string, documentId?: string) {
    return await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            updateIdInTransaction: id,
            subCollection: subCollection,
            documentId: documentId
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
}

async function getDocumentsById(ids: string[], subCollection?: string, documentId?: string) {
    return await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            methodName: "getDocumentsById",
            args: [ids],
            subCollection: subCollection,
            documentId: documentId
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
}

async function updateDocument(id: string, updateData: any, subCollection?: string, documentId?: string) {
    return await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            methodName: "updateDocument",
            args: [id, updateData],
            subCollection: subCollection,
            documentId: documentId
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
}

async function listDocuments(documentId?: string, subCollection?: string) {
    return await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            methodName: "listDocuments",
            args: [],
            subCollection: subCollection,
            documentId: documentId
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
}

async function documentExists(id: string, subCollection?: string, documentId?: string) {
    return await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            methodName: "documentExists",
            args: [id],
            subCollection: subCollection,
            documentId: documentId
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
}


async function bulkSet(data: any[], subCollection?: string, documentId?: string) {
    return await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            methodName: "bulkSet",
            args: [data],
            subCollection: subCollection,
            documentId: documentId
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
}

async function deleteDocument(id: string, subCollection?: string, documentId?: string) {
    return await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            methodName: "deleteDocument",
            args: [id],
            subCollection: subCollection,
            documentId: documentId
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
}

async function addDocument(data: any, subCollection?: string, documentId?: string) {
    return await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            methodName: "addDocument",
            args: [data],
            subCollection: subCollection,
            documentId: documentId
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
}

async function setDocument(id: string, data: any, optionalFields : string[]  = [], subCollection?: string, documentId?: string) {
    await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            methodName: "setDocument",
            args: [
                id,
                data
            ],
            optionalFields: optionalFields,
            subCollection: subCollection,
            documentId: documentId
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
    return getDocument(id)
}

async function getDocument(id: string, subCollection?: string, documentId?: string) {
    return await superagent.agent()
        .put(`${host}/${crudFunctionName}`)
        .send({
            collectionName: collectionName,
            methodName: "getDocument",
            args: [
                id
            ],
            subCollection: subCollection,
            documentId: documentId
        })
        .set("Content-Type", "application/json")
        .catch(reason => {
            return reason
        })
}


interface  EntityForCreation {
    someMandatoryAttrib : string,
    someOptionalAttrib?: string,
    someOptionalDate: Date
}

interface  Entity extends EntityForCreation{
    created: Date,
    lastModified: Date,
    id : string
    version: number
}
